# This file was automatically generated by Dist::Zilla::Plugin::Test::Pod::Coverage::TrustMe v1.0.1
use strict;
use warnings;
use Test::More;
use Test::Pod::Coverage::TrustMe;

my $config = {};
my $modules = [
  "Mojolicious::Plugin::OpenAPI::Modern",
];

plan tests => scalar @$modules;

for my $module (@$modules) {
    pod_coverage_ok($module, $config);
}

done_testing;
